@@ -1,6 +1,6 @@
 {-# LANGUAGE ScopedTypeVariables, OverloadedStrings #-}
 {-
-Copyright (C) 2006-2014 John MacFarlane <jgm@berkeley.edu>
+Copyright (C) 2006-2015 John MacFarlane <jgm@berkeley.edu>
 
 This program is free software; you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
@@ -19,7 +19,7 @@ Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 
 {- |
    Module      : Text.Pandoc.Readers.LaTeX
-   Copyright   : Copyright (C) 2006-2014 John MacFarlane
+   Copyright   : Copyright (C) 2006-2015 John MacFarlane
    License     : GNU GPL, version 2 or above
 
    Maintainer  : John MacFarlane <jgm@berkeley.edu>
@@ -42,26 +42,25 @@ import Text.Pandoc.Options
 import Text.Pandoc.Parsing hiding ((<|>), many, optional, space,
                                    mathDisplay, mathInline)
 import qualified Text.Pandoc.UTF8 as UTF8
-import Data.Char ( chr, ord )
+import Data.Char ( chr, ord, isLetter, isAlphaNum )
 import Control.Monad.Trans (lift)
 import Control.Monad
 import Text.Pandoc.Builder
-import Data.Char (isLetter, isAlphaNum)
-import Control.Applicative
-import Data.Monoid
-import Data.Maybe (fromMaybe)
+import Control.Applicative ((<|>), many, optional)
+import Data.Maybe (fromMaybe, maybeToList)
 import System.Environment (getEnv)
-import System.FilePath (replaceExtension, (</>))
-import Data.List (intercalate, intersperse)
+import System.FilePath (replaceExtension, (</>), takeExtension, addExtension)
+import Data.List (intercalate)
 import qualified Data.Map as M
 import qualified Control.Exception as E
-import System.FilePath (takeExtension, addExtension)
 import Text.Pandoc.Highlighting (fromListingsLanguage)
+import Text.Pandoc.ImageSize (numUnit, showFl)
+import Text.Pandoc.Error
 
 -- | Parse LaTeX from string and return 'Pandoc' document.
 readLaTeX :: ReaderOptions -- ^ Reader options
           -> String        -- ^ String to parse (assumes @'\n'@ line endings)
-          -> Pandoc
+          -> Either PandocError Pandoc
 readLaTeX opts = readWith parseLaTeX def{ stateOptions = opts }
 
 parseLaTeX :: LP Pandoc
@@ -73,17 +72,16 @@ parseLaTeX = do
   let (Pandoc _ bs') = doc bs
   return $ Pandoc meta bs'
 
-type LP = Parser [Char] ParserState
+type LP = Parser String ParserState
 
 anyControlSeq :: LP String
 anyControlSeq = do
   char '\\'
   next <- option '\n' anyChar
-  name <- case next of
-               '\n'           -> return ""
-               c | isLetter c -> (c:) <$> (many letter <* optional sp)
-                 | otherwise  -> return [c]
-  return name
+  case next of
+       '\n'           -> return ""
+       c | isLetter c -> (c:) <$> (many letter <* optional sp)
+         | otherwise  -> return [c]
 
 controlSeq :: String -> LP String
 controlSeq name = try $ do
@@ -102,8 +100,13 @@ dimenarg = try $ do
   return $ ch ++ num ++ dim
 
 sp :: LP ()
-sp = skipMany1 $ satisfy (\c -> c == ' ' || c == '\t')
-        <|> (try $ newline <* lookAhead anyChar <* notFollowedBy blankline)
+sp = whitespace <|> endline
+
+whitespace :: LP ()
+whitespace = skipMany1 $ satisfy (\c -> c == ' ' || c == '\t')
+
+endline :: LP ()
+endline = try (newline >> lookAhead anyChar >> notFollowedBy blankline)
 
 isLowerHex :: Char -> Bool
 isLowerHex x = x >= '0' && x <= '9' || x >= 'a' && x <= 'f'
@@ -129,7 +132,9 @@ comment = do
   return ()
 
 bgroup :: LP ()
-bgroup = () <$ char '{'
+bgroup = try $ do
+  skipMany (spaceChar <|> try (newline <* notFollowedBy blankline))
+  () <$ char '{'
      <|> () <$ controlSeq "bgroup"
      <|> () <$ controlSeq "begingroup"
 
@@ -161,51 +166,61 @@ mathInline :: LP String -> LP Inlines
 mathInline p = math <$> (try p >>= applyMacros')
 
 mathChars :: LP String
-mathChars = concat <$>
-  many (   many1 (satisfy (\c -> c /= '$' && c /='\\'))
-      <|> (\c -> ['\\',c]) <$> (try $ char '\\' *> anyChar)
-       )
+mathChars = (concat <$>) $
+  many $
+          many1 (satisfy (\c -> c /= '$' && c /='\\'))
+      <|> (\c -> ['\\',c]) <$> try (char '\\' *> anyChar)
 
 quoted' :: (Inlines -> Inlines) -> LP String -> LP () -> LP Inlines
 quoted' f starter ender = do
   startchs <- starter
-  try ((f . mconcat) <$> manyTill inline ender) <|> lit startchs
+  smart <- getOption readerSmart
+  if smart
+     then do
+       ils <- many (notFollowedBy ender >> inline)
+       (ender >> return (f (mconcat ils))) <|>
+            (<> mconcat ils) <$>
+                    lit (case startchs of
+                              "``"  -> "“"
+                              "`"   -> "‘"
+                              _     -> startchs)
+     else lit startchs
 
-double_quote :: LP Inlines
-double_quote =
-  (   quoted' doubleQuoted (try $ string "``") (void $ try $ string "''")
-  <|> quoted' doubleQuoted (string "“")        (void $ char '”')
-  -- the following is used by babel for localized quotes:
-  <|> quoted' doubleQuoted (try $ string "\"`") (void $ try $ string "\"'")
-  <|> quoted' doubleQuoted (string "\"")       (void $ char '"')
-  )
+doubleQuote :: LP Inlines
+doubleQuote = do
+  quoted' doubleQuoted (try $ string "``") (void $ try $ string "''")
+   <|> quoted' doubleQuoted (string "“")        (void $ char '”')
+   -- the following is used by babel for localized quotes:
+   <|> quoted' doubleQuoted (try $ string "\"`") (void $ try $ string "\"'")
+   <|> quoted' doubleQuoted (string "\"")       (void $ char '"')
 
-single_quote :: LP Inlines
-single_quote =
-  (  quoted' singleQuoted (string "`") (try $ char '\'' >> notFollowedBy letter)
-  <|> quoted' singleQuoted (string "‘") (try $ char '’' >> notFollowedBy letter)
-  )
+singleQuote :: LP Inlines
+singleQuote = do
+  smart <- getOption readerSmart
+  if smart
+     then quoted' singleQuoted (string "`") (try $ char '\'' >> notFollowedBy letter)
+      <|> quoted' singleQuoted (string "‘") (try $ char '’' >> notFollowedBy letter)
+     else str <$> many1 (oneOf "`\'‘’")
 
 inline :: LP Inlines
 inline = (mempty <$ comment)
-     <|> (space  <$ sp)
+     <|> (space  <$ whitespace)
+     <|> (softbreak <$ endline)
      <|> inlineText
      <|> inlineCommand
      <|> inlineEnvironment
      <|> inlineGroup
      <|> (char '-' *> option (str "-")
-           ((char '-') *> option (str "–") (str "—" <$ char '-')))
-     <|> double_quote
-     <|> single_quote
+           (char '-' *> option (str "–") (str "—" <$ char '-')))
+     <|> doubleQuote
+     <|> singleQuote
      <|> (str "”" <$ try (string "''"))
      <|> (str "”" <$ char '”')
      <|> (str "’" <$ char '\'')
      <|> (str "’" <$ char '’')
      <|> (str "\160" <$ char '~')
-     <|> (mathDisplay $ string "$$" *> mathChars <* string "$$")
-     <|> (mathInline  $ char '$' *> mathChars <* char '$')
-     <|> (superscript <$> (char '^' *> tok))
-     <|> (subscript <$> (char '_' *> tok))
+     <|> mathDisplay (string "$$" *> mathChars <* string "$$")
+     <|> mathInline  (char '$' *> mathChars <* char '$')
      <|> (guardEnabled Ext_literate_haskell *> char '|' *> doLHSverb)
      <|> (str . (:[]) <$> tildeEscape)
      <|> (str . (:[]) <$> oneOf "[]")
@@ -238,20 +253,34 @@ block = (mempty <$ comment)
 blocks :: LP Blocks
 blocks = mconcat <$> many block
 
+getRawCommand :: String -> LP String
+getRawCommand name' = do
+  rawargs <- withRaw (many (try (optional sp *> opt)) *>
+                      option "" (try (optional sp *> dimenarg)) *>
+                      many braced)
+  return $ '\\' : name' ++ snd rawargs
+
+lookupListDefault :: (Ord k) => v -> [k] -> M.Map k v -> v
+lookupListDefault d = (fromMaybe d .) . lookupList
+  where
+  lookupList l m = msum $ map (`M.lookup` m) l
+
 blockCommand :: LP Blocks
 blockCommand = try $ do
   name <- anyControlSeq
   guard $ name /= "begin" && name /= "end"
   star <- option "" (string "*" <* optional sp)
   let name' = name ++ star
-  case M.lookup name' blockCommands of
-       Just p      -> p
-       Nothing     -> case M.lookup name blockCommands of
-                           Just p    -> p
-                           Nothing   -> mzero
+  let raw = do
+        rawcommand <- getRawCommand name'
+        transformed <- applyMacros' rawcommand
+        guard $ transformed /= rawcommand
+        notFollowedBy $ parseFromString inlines transformed
+        parseFromString blocks transformed
+  lookupListDefault raw [name',name] blockCommands
 
 inBrackets :: Inlines -> Inlines
-inBrackets x = (str "[") <> x <> (str "]")
+inBrackets x = str "[" <> x <> str "]"
 
 -- eat an optional argument and one or more arguments in braces
 ignoreInlines :: String -> (String, LP Inlines)
@@ -259,19 +288,21 @@ ignoreInlines name = (name, doraw <|> (mempty <$ optargs))
   where optargs = skipopts *> skipMany (try $ optional sp *> braced)
         contseq = '\\':name
         doraw = (rawInline "latex" . (contseq ++) . snd) <$>
-                 (getOption readerParseRaw >>= guard >> (withRaw optargs))
+                 (getOption readerParseRaw >>= guard >> withRaw optargs)
 
 ignoreBlocks :: String -> (String, LP Blocks)
 ignoreBlocks name = (name, doraw <|> (mempty <$ optargs))
   where optargs = skipopts *> skipMany (try $ optional sp *> braced)
         contseq = '\\':name
         doraw = (rawBlock "latex" . (contseq ++) . snd) <$>
-                 (getOption readerParseRaw >>= guard >> (withRaw optargs))
+                 (getOption readerParseRaw >>= guard >> withRaw optargs)
 
 blockCommands :: M.Map String (LP Blocks)
 blockCommands = M.fromList $
   [ ("par", mempty <$ skipopts)
-  , ("title", mempty <$ (skipopts *> tok >>= addMeta "title"))
+  , ("title", mempty <$ (skipopts *>
+                          (grouped inline >>= addMeta "title")
+                      <|> (grouped block >>= addMeta "title")))
   , ("subtitle", mempty <$ (skipopts *> tok >>= addMeta "subtitle"))
   , ("author", mempty <$ (skipopts *> authors))
   -- -- in letter class, temp. store address & sig as title, author
@@ -302,7 +333,7 @@ blockCommands = M.fromList $
   --
   , ("hrule", pure horizontalRule)
   , ("rule", skipopts *> tok *> tok *> pure horizontalRule)
-  , ("item", skipopts *> loose_item)
+  , ("item", skipopts *> looseItem)
   , ("documentclass", skipopts *> braced *> preamble)
   , ("centerline", (para . trimInlines) <$> (skipopts *> tok))
   , ("caption", skipopts *> setCaption)
@@ -328,6 +359,7 @@ blockCommands = M.fromList $
   , "hyperdef"
   , "markboth", "markright", "markleft"
   , "hspace", "vspace"
+  , "newpage"
   ]
 
 addMeta :: ToMetaValue a => String -> a -> LP ()
@@ -341,7 +373,7 @@ setCaption :: LP Blocks
 setCaption = do
   ils <- tok
   mblabel <- option Nothing $
-               try $ spaces >> controlSeq "label" >> (Just <$> tok)
+               try $ spaces' >> controlSeq "label" >> (Just <$> tok)
   let ils' = case mblabel of
                   Just lab -> ils <> spanWith
                                 ("",[],[("data-label", stringify lab)]) mempty
@@ -369,7 +401,7 @@ section (ident, classes, kvs) lvl = do
   let lvl' = if hasChapters then lvl + 1 else lvl
   skipopts
   contents <- grouped inline
-  lab <- option ident $ try (spaces >> controlSeq "label" >> spaces >> braced)
+  lab <- option ident $ try (spaces' >> controlSeq "label" >> spaces' >> braced)
   attr' <- registerHeader (lab, classes, kvs) contents
   return $ headerWith attr' lvl' contents
 
@@ -390,17 +422,14 @@ inlineCommand = try $ do
            else if parseRaw
                    then return $ rawInline "latex" rawcommand
                    else return mempty
-  case M.lookup name' inlineCommands of
-       Just p      -> p <|> raw
-       Nothing     -> case M.lookup name inlineCommands of
-                           Just p    -> p <|> raw
-                           Nothing   -> raw
+  lookupListDefault mzero [name',name] inlineCommands
+    <|> raw
 
 unlessParseRaw :: LP ()
 unlessParseRaw = getOption readerParseRaw >>= guard . not
 
 isBlockCommand :: String -> Bool
-isBlockCommand s = maybe False (const True) $ M.lookup s blockCommands
+isBlockCommand s = s `M.member` blockCommands
 
 
 inlineEnvironments :: M.Map String (LP Inlines)
@@ -446,7 +475,7 @@ inlineCommands = M.fromList $
   , ("cref", unlessParseRaw >> (inBrackets <$> tok))  -- from cleveref.sty
   , ("(", mathInline $ manyTill anyChar (try $ string "\\)"))
   , ("[", mathDisplay $ manyTill anyChar (try $ string "\\]"))
-  , ("ensuremath", mathInline $ braced)
+  , ("ensuremath", mathInline braced)
   , ("texorpdfstring", (\_ x -> x) <$> tok <*> tok)
   , ("P", lit "¶")
   , ("S", lit "§")
@@ -495,7 +524,7 @@ inlineCommands = M.fromList $
   , ("v", option (str "v") $ try $ tok >>= accent hacek)
   , ("u", option (str "u") $ try $ tok >>= accent breve)
   , ("i", lit "i")
-  , ("\\", linebreak <$ (optional (bracketed inline) *> optional sp))
+  , ("\\", linebreak <$ (optional (bracketed inline) *> spaces'))
   , (",", pure mempty)
   , ("@", pure mempty)
   , (" ", lit "\160")
@@ -508,7 +537,7 @@ inlineCommands = M.fromList $
   , ("thanks", (note . mconcat) <$> (char '{' *> manyTill block (char '}')))
   , ("footnote", (note . mconcat) <$> (char '{' *> manyTill block (char '}')))
   , ("verb", doverb)
-  , ("lstinline", doverb)
+  , ("lstinline", skipopts *> doverb)
   , ("Verb", doverb)
   , ("texttt", (code . stringify . toList) <$> tok)
   , ("url", (unescapeURL <$> braced) >>= \url ->
@@ -516,9 +545,12 @@ inlineCommands = M.fromList $
   , ("href", (unescapeURL <$> braced <* optional sp) >>= \url ->
        tok >>= \lab ->
          pure (link url "" lab))
-  , ("includegraphics", skipopts *> (unescapeURL <$> braced) >>= mkImage)
+  , ("includegraphics", do options <- option [] keyvals
+                           src <- unescapeURL <$> braced
+                           mkImage options src)
   , ("enquote", enquote)
   , ("cite", citation "cite" AuthorInText False)
+  , ("Cite", citation "cite" AuthorInText False)
   , ("citep", citation "citep" NormalCitation False)
   , ("citep*", citation "citep*" NormalCitation False)
   , ("citeal", citation "citeal" NormalCitation False)
@@ -577,14 +609,19 @@ inlineCommands = M.fromList $
   -- in which case they will appear as raw latex blocks:
   [ "index" ]
 
-mkImage :: String -> LP Inlines
-mkImage src = do
+mkImage :: [(String, String)] -> String -> LP Inlines
+mkImage options src = do
+   let replaceTextwidth (k,v) = case numUnit v of
+                                  Just (num, "\\textwidth") -> (k, showFl (num * 100) ++ "%")
+                                  _ -> (k, v)
+   let kvs = map replaceTextwidth $ filter (\(k,_) -> k `elem` ["width", "height"]) options
+   let attr = ("",[], kvs)
    let alt = str "image"
    case takeExtension src of
         "" -> do
               defaultExt <- getOption readerDefaultImageExtension
-              return $ image (addExtension src defaultExt) "" alt
-        _  -> return $ image src "" alt
+              return $ imageWith attr (addExtension src defaultExt) "" alt
+        _  -> return $ imageWith attr src "" alt
 
 inNote :: Inlines -> Inlines
 inNote ils =
@@ -618,7 +655,7 @@ lit = pure . str
 accent :: (Char -> String) -> Inlines -> LP Inlines
 accent f ils =
   case toList ils of
-       (Str (x:xs) : ys) -> return $ fromList $ (Str (f x ++ xs) : ys)
+       (Str (x:xs) : ys) -> return $ fromList (Str (f x ++ xs) : ys)
        []                -> mzero
        _                 -> return ils
 
@@ -807,35 +844,32 @@ breve 'u' = "ŭ"
 breve c   = [c]
 
 tok :: LP Inlines
-tok = try $ grouped inline <|> inlineCommand <|> str <$> (count 1 $ inlineChar)
+tok = try $ grouped inline <|> inlineCommand <|> str <$> count 1 inlineChar
 
 opt :: LP Inlines
-opt = bracketed inline <* optional sp
+opt = bracketed inline
 
 skipopts :: LP ()
-skipopts = skipMany opt
+skipopts = skipMany (opt *> optional sp)
 
 inlineText :: LP Inlines
 inlineText = str <$> many1 inlineChar
 
 inlineChar :: LP Char
-inlineChar = noneOf "\\$%^_&~#{}^'`\"‘’“”-[] \t\n"
+inlineChar = noneOf "\\$%&~#{}^'`\"‘’“”-[] \t\n"
 
 environment :: LP Blocks
 environment = do
   controlSeq "begin"
   name <- braced
-  case M.lookup name environments of
-       Just p      -> p <|> rawEnv name
-       Nothing     -> rawEnv name
+  M.findWithDefault mzero name environments
+    <|> rawEnv name
 
 inlineEnvironment :: LP Inlines
 inlineEnvironment = try $ do
   controlSeq "begin"
   name <- braced
-  case M.lookup name inlineEnvironments of
-       Just p      -> p
-       Nothing     -> mzero
+  M.findWithDefault mzero name inlineEnvironments
 
 rawEnv :: String -> LP Blocks
 rawEnv name = do
@@ -848,15 +882,11 @@ rawEnv name = do
 
 ----
 
-type IncludeParser = ParserT [Char] [String] IO String
+type IncludeParser = ParserT String [String] IO String
 
 -- | Replace "include" commands with file contents.
-handleIncludes :: String -> IO String
-handleIncludes s = do
-  res <- runParserT includeParser' [] "input" s
-  case res of
-       Right s'    -> return s'
-       Left e      -> error $ show e
+handleIncludes :: String -> IO (Either PandocError String)
+handleIncludes s =  mapLeft (ParsecError s) <$> runParserT includeParser' [] "input" s
 
 includeParser' :: IncludeParser
 includeParser' =
@@ -886,7 +916,7 @@ verbatimEnv' = fmap snd <$>
              string "\\begin"
              name <- braced'
              guard $ name `elem` ["verbatim", "Verbatim", "lstlisting",
-                                  "minted", "alltt"]
+                                  "minted", "alltt", "comment"]
              manyTill anyChar (try $ string $ "\\end{" ++ name ++ "}")
 
 blob' :: IncludeParser
@@ -898,6 +928,12 @@ backslash' = string "\\"
 braced' :: IncludeParser
 braced' = try $ char '{' *> manyTill (satisfy (/='}')) (char '}')
 
+maybeAddExtension :: String -> FilePath -> FilePath
+maybeAddExtension ext fp =
+  if null (takeExtension fp)
+     then addExtension fp ext
+     else fp
+
 include' :: IncludeParser
 include' = do
   fs' <- try $ do
@@ -906,11 +942,11 @@ include' = do
                   <|> try (string "input")
                   <|> string "usepackage"
               -- skip options
-              skipMany $ try $ char '[' *> (manyTill anyChar (char ']'))
+              skipMany $ try $ char '[' *> manyTill anyChar (char ']')
               fs <- (map trim . splitBy (==',')) <$> braced'
               return $ if name == "usepackage"
-                          then map (flip replaceExtension ".sty") fs
-                          else map (flip replaceExtension ".tex") fs
+                          then map (maybeAddExtension ".sty") fs
+                          else map (maybeAddExtension ".tex") fs
   pos <- getPosition
   containers <- getState
   let fn = case containers of
@@ -966,7 +1002,7 @@ readFileFromDirs (d:ds) f =
 keyval :: LP (String, String)
 keyval = try $ do
   key <- many1 alphaNum
-  val <- option "" $ char '=' >> many1 alphaNum
+  val <- option "" $ char '=' >> many1 (alphaNum <|> char '.' <|> char '\\')
   skipMany spaceChar
   optional (char ',')
   skipMany spaceChar
@@ -979,57 +1015,62 @@ keyvals = try $ char '[' *> manyTill keyval (char ']')
 alltt :: String -> LP Blocks
 alltt t = walk strToCode <$> parseFromString blocks
   (substitute " " "\\ " $ substitute "%" "\\%" $
-   concat $ intersperse "\\\\\n" $ lines t)
+   intercalate "\\\\\n" $ lines t)
   where strToCode (Str s) = Code nullAttr s
         strToCode x       = x
 
-rawLaTeXBlock :: Parser [Char] ParserState String
+rawLaTeXBlock :: LP String
 rawLaTeXBlock = snd <$> try (withRaw (environment <|> blockCommand))
 
-rawLaTeXInline :: Parser [Char] ParserState Inline
+rawLaTeXInline :: LP Inline
 rawLaTeXInline = do
   raw <- (snd <$> withRaw inlineCommand) <|> (snd <$> withRaw blockCommand)
   RawInline "latex" <$> applyMacros' raw
 
 addImageCaption :: Blocks -> LP Blocks
 addImageCaption = walkM go
-  where go (Image alt (src,tit)) = do
+  where go (Image attr alt (src,tit)) = do
           mbcapt <- stateCaption <$> getState
-          case mbcapt of
-               Just ils -> return (Image (toList ils) (src, "fig:"))
-               Nothing  -> return (Image alt (src,tit))
+          return $ case mbcapt of
+               Just ils -> Image attr (toList ils) (src, "fig:")
+               Nothing  -> Image attr alt (src,tit)
         go x = return x
 
 addTableCaption :: Blocks -> LP Blocks
 addTableCaption = walkM go
   where go (Table c als ws hs rs) = do
           mbcapt <- stateCaption <$> getState
-          case mbcapt of
-               Just ils -> return (Table (toList ils) als ws hs rs)
-               Nothing  -> return (Table c als ws hs rs)
+          return $ case mbcapt of
+               Just ils -> Table (toList ils) als ws hs rs
+               Nothing  -> Table c als ws hs rs
         go x = return x
 
 environments :: M.Map String (LP Blocks)
 environments = M.fromList
   [ ("document", env "document" blocks <* skipMany anyChar)
-  , ("letter", env "letter" letter_contents)
+  , ("abstract", mempty <$ (env "abstract" blocks >>= addMeta "abstract"))
+  , ("letter", env "letter" letterContents)
   , ("figure", env "figure" $
          resetCaption *> skipopts *> blocks >>= addImageCaption)
   , ("center", env "center" blocks)
+  , ("longtable",  env "longtable" $
+         resetCaption *> skipopts *> blocks >>= addTableCaption)
   , ("table",  env "table" $
          resetCaption *> skipopts *> blocks >>= addTableCaption)
-  , ("tabular", env "tabular" simpTable)
+  , ("tabular*", env "tabular" $ simpTable True)
+  , ("tabular", env "tabular"  $ simpTable False)
   , ("quote", blockQuote <$> env "quote" blocks)
   , ("quotation", blockQuote <$> env "quotation" blocks)
   , ("verse", blockQuote <$> env "verse" blocks)
   , ("itemize", bulletList <$> listenv "itemize" (many item))
   , ("description", definitionList <$> listenv "description" (many descItem))
-  , ("enumerate", ordered_list)
+  , ("enumerate", orderedList')
   , ("alltt", alltt =<< verbEnv "alltt")
   , ("code", guardEnabled Ext_literate_haskell *>
       (codeBlockWith ("",["sourceCode","literate","haskell"],[]) <$>
         verbEnv "code"))
-  , ("verbatim", codeBlock <$> (verbEnv "verbatim"))
+  , ("comment", mempty <$ verbEnv "comment")
+  , ("verbatim", codeBlock <$> verbEnv "verbatim")
   , ("Verbatim",   do options <- option [] keyvals
                       let kvs = [ (if k == "firstnumber"
                                       then "startFrom"
@@ -1037,17 +1078,17 @@ environments = M.fromList
                       let classes = [ "numberLines" |
                                       lookup "numbers" options == Just "left" ]
                       let attr = ("",classes,kvs)
-                      codeBlockWith attr <$> (verbEnv "Verbatim"))
+                      codeBlockWith attr <$> verbEnv "Verbatim")
   , ("lstlisting", do options <- option [] keyvals
                       let kvs = [ (if k == "firstnumber"
                                       then "startFrom"
                                       else k, v) | (k,v) <- options ]
                       let classes = [ "numberLines" |
                                       lookup "numbers" options == Just "left" ]
-                                 ++ maybe [] (:[]) (lookup "language" options
+                                 ++ maybeToList (lookup "language" options
                                          >>= fromListingsLanguage)
                       let attr = (fromMaybe "" (lookup "label" options),classes,kvs)
-                      codeBlockWith attr <$> (verbEnv "lstlisting"))
+                      codeBlockWith attr <$> verbEnv "lstlisting")
   , ("minted",     do options <- option [] keyvals
                       lang <- grouped (many1 $ satisfy (/='}'))
                       let kvs = [ (if k == "firstnumber"
@@ -1057,7 +1098,7 @@ environments = M.fromList
                                     [ "numberLines" |
                                       lookup "linenos" options == Just "true" ]
                       let attr = ("",classes,kvs)
-                      codeBlockWith attr <$> (verbEnv "minted"))
+                      codeBlockWith attr <$> verbEnv "minted")
   , ("obeylines", parseFromString
                   (para . trimInlines . mconcat <$> many inline) =<<
                   intercalate "\\\\\n" . lines <$> verbEnv "obeylines")
@@ -1076,8 +1117,8 @@ environments = M.fromList
   , ("alignat*", mathEnv para (Just "aligned") "alignat*")
   ]
 
-letter_contents :: LP Blocks
-letter_contents = do
+letterContents :: LP Blocks
+letterContents = do
   bs <- blocks
   st <- getState
   -- add signature (author) and address (title)
@@ -1104,8 +1145,8 @@ closing = do
 item :: LP Blocks
 item = blocks *> controlSeq "item" *> skipopts *> blocks
 
-loose_item :: LP Blocks
-loose_item = do
+looseItem :: LP Blocks
+looseItem = do
   ctx <- stateParserContext `fmap` getState
   if ctx == ListItemState
      then mzero
@@ -1148,8 +1189,8 @@ verbEnv name = do
   res <- manyTill anyChar endEnv
   return $ stripTrailingNewlines res
 
-ordered_list :: LP Blocks
-ordered_list = do
+orderedList' :: LP Blocks
+orderedList' = do
   optional sp
   (_, style, delim) <- option (1, DefaultStyle, DefaultDelim) $
                               try $ char '[' *> anyOrderedListMarker <* char ']'
@@ -1161,7 +1202,7 @@ ordered_list = do
                                optional sp
                                num <- grouped (many1 digit)
                                spaces
-                               return $ (read num + 1 :: Int)
+                               return (read num + 1 :: Int)
   bs <- listenv "enumerate" (many item)
   return $ orderedListWith (start, style, delim) bs
 
@@ -1175,14 +1216,14 @@ paragraph = do
 preamble :: LP Blocks
 preamble = mempty <$> manyTill preambleBlock beginDoc
   where beginDoc = lookAhead $ try $ controlSeq "begin" *> string "{document}"
-        preambleBlock =  (void comment)
-                     <|> (void sp)
-                     <|> (void blanklines)
-                     <|> (void macro)
-                     <|> (void blockCommand)
-                     <|> (void anyControlSeq)
-                     <|> (void braced)
-                     <|> (void anyChar)
+        preambleBlock =  void comment
+                     <|> void sp
+                     <|> void blanklines
+                     <|> void macro
+                     <|> void blockCommand
+                     <|> void anyControlSeq
+                     <|> void braced
+                     <|> void anyChar
 
 -------
 
@@ -1258,7 +1299,7 @@ complexNatbibCitation mode = try $ do
                    suff  <- ils
                    skipSpaces
                    optional $ char ';'
-                   return $ addPrefix pref $ addSuffix suff $ cits'
+                   return $ addPrefix pref $ addSuffix suff cits'
   (c:cits, raw) <- withRaw $ grouped parseOne
   return $ cite (c{ citationMode = mode }:cits)
            (rawInline "latex" $ "\\citetext" ++ raw)
@@ -1268,7 +1309,7 @@ complexNatbibCitation mode = try $ do
 parseAligns :: LP [Alignment]
 parseAligns = try $ do
   char '{'
-  let maybeBar = skipMany $ sp <|> () <$ char '|' <|> () <$ try (string "@{}")
+  let maybeBar = skipMany $ sp <|> () <$ char '|' <|> () <$ (char '@' >> braced)
   maybeBar
   let cAlign = AlignCenter <$ char 'c'
   let lAlign = AlignLeft <$ char 'l'
@@ -1282,13 +1323,22 @@ parseAligns = try $ do
   return aligns'
 
 hline :: LP ()
-hline = () <$ (try $ spaces >> controlSeq "hline")
+hline = try $ do
+  spaces'
+  controlSeq "hline" <|>
+    -- booktabs rules:
+    controlSeq "toprule" <|>
+    controlSeq "bottomrule" <|>
+    controlSeq "midrule"
+  spaces'
+  optional $ bracketed (many1 (satisfy (/=']')))
+  return ()
 
 lbreak :: LP ()
-lbreak = () <$ (try $ spaces *> controlSeq "\\")
+lbreak = () <$ try (spaces' *> controlSeq "\\" <* spaces')
 
 amp :: LP ()
-amp = () <$ (try $ spaces *> char '&')
+amp = () <$ try (spaces' *> char '&')
 
 parseTableRow :: Int  -- ^ number of columns
               -> LP [Blocks]
@@ -1301,19 +1351,22 @@ parseTableRow cols = try $ do
   guard $ cells' /= [mempty]
   -- note:  a & b in a three-column table leaves an empty 3rd cell:
   let cells'' = cells' ++ replicate (cols - numcells) mempty
-  spaces
+  spaces'
   return cells''
 
-simpTable :: LP Blocks
-simpTable = try $ do
-  spaces
+spaces' :: LP ()
+spaces' = spaces *> skipMany (comment *> spaces)
+
+simpTable :: Bool -> LP Blocks
+simpTable hasWidthParameter = try $ do
+  when hasWidthParameter $ () <$ (spaces' >> tok)
+  skipopts
   aligns <- parseAligns
   let cols = length aligns
   optional hline
   header' <- option [] $ try (parseTableRow cols <* lbreak <* hline)
   rows <- sepEndBy (parseTableRow cols) (lbreak <* optional hline)
-  spaces
-  skipMany (comment *> spaces)
+  spaces'
   let header'' = if null header'
                     then replicate cols mempty
                     else header'
