@@ -7,10 +7,10 @@ import Tests.Helpers
 import Tests.Arbitrary()
 import Text.Pandoc.Builder
 import Text.Pandoc
-import Data.Monoid (mempty)
+import Text.Pandoc.Error
 
 rst :: String -> Pandoc
-rst = readRST def{ readerStandalone = True }
+rst = handleError . readRST def{ readerStandalone = True }
 
 infix 4 =:
 (=:) :: ToString c
@@ -21,7 +21,8 @@ tests :: [Test]
 tests = [ "line block with blank line" =:
           "| a\n|\n|  b" =?> para (str "a") <>
                              para (str "\160b")
-        , "field list" =: unlines
+        , testGroup "field list"
+          [ "general" =: unlines
              [ "para"
              , ""
              , ":Hostname: media08"
@@ -35,17 +36,17 @@ tests = [ "line block with blank line" =:
              , ":Parameter i: integer"
              , ":Final: item"
              , "  on two lines" ]
-           =?> ( doc
-               $ para "para" <>
-                 definitionList [ (str "Hostname", [para "media08"])
-                                , (str "IP address", [para "10.0.0.19"])
-                                , (str "Size", [para "3ru"])
-                                , (str "Version", [para "1"])
-                                , (str "Indentation", [para "Since the field marker may be quite long, the second and subsequent lines of the field body do not have to line up with the first line, but they must be indented relative to the field name marker, and they must line up with each other."])
-                                , (str "Parameter i", [para "integer"])
-                                , (str "Final", [para "item on two lines"])
-                              ])
-        , "initial field list" =: unlines
+             =?> ( doc
+                 $ para "para" <>
+                   definitionList [ (str "Hostname", [para "media08"])
+                                  , (text "IP address", [para "10.0.0.19"])
+                                  , (str "Size", [para "3ru"])
+                                  , (str "Version", [para "1"])
+                                  , (str "Indentation", [para "Since the field marker may be quite long, the second\nand subsequent lines of the field body do not have to line up\nwith the first line, but they must be indented relative to the\nfield name marker, and they must line up with each other."])
+                                  , (text "Parameter i", [para "integer"])
+                                  , (str "Final", [para "item\non two lines"])
+                                  ])
+          , "metadata" =: unlines
              [ "====="
              , "Title"
              , "====="
@@ -55,18 +56,44 @@ tests = [ "line block with blank line" =:
              , ""
              , ":Version: 1"
              ]
-           =?> ( setMeta "version" (para "1")
-               $ setMeta "title" ("Title" :: Inlines)
-               $ setMeta "subtitle" ("Subtitle" :: Inlines)
-               $ doc mempty )
+             =?> ( setMeta "version" (para "1")
+                 $ setMeta "title" ("Title" :: Inlines)
+                 $ setMeta "subtitle" ("Subtitle" :: Inlines)
+                 $ doc mempty )
+          , "with inline markup" =: unlines
+             [ ":*Date*: today"
+             , ""
+             , ".."
+             , ""
+             , ":*one*: emphasis"
+             , ":two_: reference"
+             , ":`three`_: another one"
+             , ":``four``: literal"
+             , ""
+             , ".. _two: http://example.com"
+             , ".. _three: http://example.org"
+             ]
+             =?> ( setMeta "date" (str "today")
+                 $ doc
+                 $ definitionList [ (emph "one", [para "emphasis"])
+                                  , (link "http://example.com" "" "two", [para "reference"])
+                                  , (link "http://example.org" "" "three", [para "another one"])
+                                  , (code "four", [para "literal"])
+                                  ])
+          ]
         , "URLs with following punctuation" =:
           ("http://google.com, http://yahoo.com; http://foo.bar.baz.\n" ++
            "http://foo.bar/baz_(bam) (http://foo.bar)") =?>
           para (link "http://google.com" "" "http://google.com" <> ", " <>
                 link "http://yahoo.com" "" "http://yahoo.com" <> "; " <>
                 link "http://foo.bar.baz" "" "http://foo.bar.baz" <> ". " <>
+                softbreak <>
                 link "http://foo.bar/baz_(bam)" "" "http://foo.bar/baz_(bam)"
                 <> " (" <> link "http://foo.bar" "" "http://foo.bar" <> ")")
+        , "Reference names with special characters" =:
+                   ("A-1-B_2_C:3:D+4+E.5.F_\n\n" ++
+                   ".. _A-1-B_2_C:3:D+4+E.5.F: https://example.com\n") =?>
+                   para (link "https://example.com" "" "A-1-B_2_C:3:D+4+E.5.F")
         , testGroup "literal / line / code blocks"
           [ "indented literal block" =: unlines
             [ "::"
