@@ -8,7 +8,7 @@ import Tests.Helpers
 import Tests.Arbitrary()
 
 html :: (ToString a, ToPandoc a) => a -> String
-html = writeHtmlString def{ writerWrapText = False } . toPandoc
+html = writeHtmlString def{ writerWrapText = WrapNone } . toPandoc
 
 {-
   "my test" =: X =?> Y
