@@ -8,8 +8,10 @@ import qualified Tests.Old
 import qualified Tests.Readers.LaTeX
 import qualified Tests.Readers.Markdown
 import qualified Tests.Readers.Org
+import qualified Tests.Readers.HTML
 import qualified Tests.Readers.RST
 import qualified Tests.Readers.Docx
+import qualified Tests.Readers.Odt
 import qualified Tests.Readers.Txt2Tags
 import qualified Tests.Readers.EPUB
 import qualified Tests.Writers.ConTeXt
@@ -20,6 +22,8 @@ import qualified Tests.Writers.Native
 import qualified Tests.Writers.Markdown
 import qualified Tests.Writers.Plain
 import qualified Tests.Writers.AsciiDoc
+import qualified Tests.Writers.Docx
+import qualified Tests.Writers.RST
 import qualified Tests.Shared
 import qualified Tests.Walk
 import Text.Pandoc.Shared (inDirectory)
@@ -38,13 +42,17 @@ tests = [ testGroup "Old" Tests.Old.tests
           , testGroup "Markdown" Tests.Writers.Markdown.tests
           , testGroup "Plain" Tests.Writers.Plain.tests
           , testGroup "AsciiDoc" Tests.Writers.AsciiDoc.tests
+          , testGroup "Docx" Tests.Writers.Docx.tests
+          , testGroup "RST" Tests.Writers.RST.tests
           ]
         , testGroup "Readers"
           [ testGroup "LaTeX" Tests.Readers.LaTeX.tests
           , testGroup "Markdown" Tests.Readers.Markdown.tests
+          , testGroup "HTML" Tests.Readers.HTML.tests
           , testGroup "Org" Tests.Readers.Org.tests
           , testGroup "RST" Tests.Readers.RST.tests
           , testGroup "Docx" Tests.Readers.Docx.tests
+          , testGroup "Odt" Tests.Readers.Odt.tests
           , testGroup "Txt2Tags" Tests.Readers.Txt2Tags.tests
           , testGroup "EPUB" Tests.Readers.EPUB.tests
           ]
